"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.NodeUnilateral = void 0;
var linked_list_1 = require("../utilities/linked-list");
var NodeUnilateral = /** @class */ (function () {
    function NodeUnilateral(value, next) {
        this.value = value;
        this.next = next;
    }
    return NodeUnilateral;
}());
exports.NodeUnilateral = NodeUnilateral;
var SinglyLinkedList = /** @class */ (function () {
    /**
     * Creates a new instance of a singly-linked list.
     * @param values The inital values to populate list with. If omitted, list begins empty.
     */
    function SinglyLinkedList(values) {
        this._headNode = null;
        this._tailNode = null;
        this._length = 0;
        if (values !== undefined && values.length > 0) {
            (0, linked_list_1.listInit)(this, values);
        }
    }
    /**
     * Inserts value at end of list. Time complexity = O(1).
     * @param value The value to append.
     * @returns The list instance - this.
     */
    SinglyLinkedList.prototype.append = function (value) {
        var newNode = new NodeUnilateral(value, null);
        if (this._length === 0) {
            this._headNode = newNode;
        }
        if (this._tailNode !== null) {
            this._tailNode.next = newNode;
        }
        this._tailNode = newNode;
        ++this._length;
        return this;
    };
    /**
     * Inserts value at beginning of list. Time complexity = O(1).
     * @param value The value to prepend.
     * @returns The list instance - this.
     */
    SinglyLinkedList.prototype.prepend = function (value) {
        var newNode = new NodeUnilateral(value, this._headNode);
        this._headNode = newNode;
        if (this._length === 0) {
            this._tailNode = newNode;
        }
        ++this._length;
        return this;
    };
    /**
     * Inserts item at specified index. Time complexity = O(n) where n is number of items in list.
     * @param value The item to insert.
     * @param index The index at which to insert item.
     * @returns The list instance - this.
     */
    SinglyLinkedList.prototype.insert = function (value, index) {
        if (index < 0 || index > this._length) {
            throw new RangeError("cannot insert at out of bounds index " + index + ", index must be between 0 and " + this._length);
        }
        if (index === 0) {
            return this.prepend(value);
        }
        if (index === this._length) {
            return this.append(value);
        }
        var trailingNode = this._getNode(index - 1);
        var newNode = new NodeUnilateral(value, trailingNode.next);
        trailingNode.next = newNode;
        ++this._length;
        return this;
    };
    /**
     * Removes an item at specified index. Throws error is list is empty. Time complexity = O(n) where n is number of items in list.
     * @param index The index of item to remove.
     * @returns The removed item.
     */
    SinglyLinkedList.prototype.remove = function (index) {
        if (this._length === 0) {
            throw new Error("cannot remove from an empty list");
        }
        if (index === 0) {
            var nodeToRemove_1 = this._headNode;
            var trailingNode = nodeToRemove_1.next;
            if (trailingNode === null) {
                this._headNode = null;
                this._tailNode = null;
            }
            else {
                this._headNode = trailingNode;
            }
            --this._length;
            return nodeToRemove_1.value;
        }
        var leadingNode = this._getNode(index - 1);
        var nodeToRemove = leadingNode.next;
        leadingNode.next = nodeToRemove !== null ? nodeToRemove.next : null;
        if (this._length === 2) {
            this._tailNode = this._headNode;
        }
        else if (index === this._length - 1) {
            this._tailNode = leadingNode;
        }
        --this._length;
        return nodeToRemove === null || nodeToRemove === void 0 ? void 0 : nodeToRemove.value;
    };
    /**
     * Gets item at specified index. Time complexity = O(n) where n is number of items in list.
     * @param index The index of item to retrieve.
     * @returns The item at specified index.
     */
    SinglyLinkedList.prototype.get = function (index) {
        return (0, linked_list_1.listGet)(this, index);
    };
    SinglyLinkedList.prototype._getNode = function (index) {
        if (this._length === 0) {
            return null;
        }
        var currentNode = this._headNode;
        for (var i = 0; i < index; ++i) {
            if (currentNode === null) {
                (0, linked_list_1.listThrowTraversalError)();
            }
            else {
                currentNode = currentNode.next;
            }
        }
        return currentNode;
    };
    /**
     * @returns An array containing currently held items, ordered from head to tail. Time complexity = O(n) where n is number of items in list.
     */
    SinglyLinkedList.prototype.getItems = function () {
        var values = [];
        this.forEach(function (val) { return values.push(val); });
        return values;
    };
    /**
     * Executes a callback function for each item in list. Time complexity = O(n) where n is number of items in list.
     * @param callback The function to execute for each item in list.
     * @returns The list instance - this.
     */
    SinglyLinkedList.prototype.forEach = function (callback) {
        var currentNode = this._headNode;
        var index = 0;
        while (currentNode !== null) {
            callback(currentNode.value, index);
            currentNode = currentNode.next;
            ++index;
        }
        return this;
    };
    /**
     * Executes a callback function that replaces each item in list with result. Time complexity = O(n) where n is number of items in list.
     * @param callback The function to execute for each item in list.
     * @returns The list instance - this.
     */
    SinglyLinkedList.prototype.forEachMutate = function (callback) {
        var currentNode = this._headNode;
        var index = 0;
        while (currentNode !== null) {
            currentNode.value = callback(currentNode.value, index);
            currentNode = currentNode.next;
            ++index;
        }
        return this;
    };
    /**
     * Reverses items in list: head becomes tail and vice versa. Time complexity = O(n) where n is number of items in list.
     * @returns The list instance - this.
     */
    SinglyLinkedList.prototype.reverse = function () {
        if (this._length < 2) {
            return this;
        }
        var first = this._headNode;
        this._tailNode = this._headNode;
        var second = this._headNode.next;
        while (second !== null) {
            var third = second.next;
            second.next = first;
            first = second;
            second = third;
        }
        this._tailNode.next = null;
        this._headNode = first;
        return this;
    };
    /**
     * Removes all items from list. Time complexity = O(1).
     * @returns The list instance - this.
     */
    SinglyLinkedList.prototype.clear = function () {
        this._headNode = null;
        this._tailNode = null;
        this._length = 0;
        return this;
    };
    /**
     * Time complexity = O(1).
     * @returns True if list is empty, false otherwise.
     */
    SinglyLinkedList.prototype.isEmpty = function () {
        return this._length === 0;
    };
    /**
     * Time complexity = O(1).
     * @returns The current number of items stored.
     */
    SinglyLinkedList.prototype.length = function () {
        return this._length;
    };
    /**
     * Time complexity = O(1).
     * @returns The current head value, or null if list is empty.
     */
    SinglyLinkedList.prototype.head = function () {
        var _a;
        return ((_a = this._headNode) === null || _a === void 0 ? void 0 : _a.value) || null;
    };
    /**
     * Time complexity = O(1).
     * @returns The current head node, or null if list is empty.
     */
    SinglyLinkedList.prototype.headNode = function () {
        return this._headNode;
    };
    /**
     * Time complexity = O(1).
     * @returns The current tail value, or null if list is empty.
     */
    SinglyLinkedList.prototype.tail = function () {
        var _a;
        return ((_a = this._tailNode) === null || _a === void 0 ? void 0 : _a.value) || null;
    };
    /**
     * Time complexity = O(1).
     * @returns The current tail node, or null if list is empty.
     */
    SinglyLinkedList.prototype.tailNode = function () {
        return this._tailNode;
    };
    return SinglyLinkedList;
}());
exports.default = SinglyLinkedList;
